-- Poetry Augmented Programming
-- A really silly successor to Fiction Augmented Programming
-- Kinda like MIPS assembly encoded in poetry (or prose if you really want to -- harder)

--options
strict = false
strict_no_numbers = false

--interpreter storage stuff
registers = {0, 1, 0, 0, 0, 0, 0, 0}
stack = {}
stack_ptr = 0
heap = {}

--the decoded instruction store
instructions = {}
instr_ptr = 0

--print out a table
function printTable(t)
  print ""
  for k,v in pairs(t) do
    print(k,v)
  end
end

--function to decode a word
function decode_word(word)
  local vowels = 0
  if strict_no_numbers == false and tonumber(word) ~= nil then return tonumber(word) end
  for c in string.gmatch(word, "[aeiouAEIOU]") do
    vowels = vowels + 1
  end
  return vowels
end

--function to process a line
function process_line(line)
  local instr = {}
  for w in string.gmatch(line, "[%w']+") do
    instr[#instr+1] = decode_word(w)
  end
  return instr
end

--function to process a file
function process_file(f)
  for line in io.lines(f) do
    local instr = process_line(line)
    instr_ptr = instr_ptr + 1
    instructions[instr_ptr] = instr
  end
  instr_ptr = 1
end

function pop()
  stack[stack_ptr] = nil
  stack_ptr = stack_ptr - 1
end

function push(value)
  stack_ptr = stack_ptr + 1
  stack[stack_ptr] = value
end

--function to get a value from the heap (wraps around the sparse table)
function get_heap(addr)
  if heap[addr] then
    return heap[addr]
  else
    return 0
  end
end

--interpreter goes here
function interpret()
  while (instructions[instr_ptr] ~= nil) do
    local instr = instructions[instr_ptr]
    
    if instr[1] == 1 then -- stack instructions
      if instr[2] == 1 then -- pop
        if stack_ptr ~= 0 then
          pop()
        else
          print("Line: "..instr_ptr..", ERROR: Can't POP. Empty stack.")
          if strict then return end
        end
      elseif instr[2] == 2 then -- push
        if instr[3] < 1 or instr[3] > 8 then
          print("Line: "..instr_ptr..", ERROR: Can't PUSH. Illegal register.")
          if strict then return end
        else
          push(registers[instr[3]])
        end
      elseif instr[2] == 3 then -- copy
        if instr[3] < 3 or instr[3] > 8 then
          print("Line: "..instr_ptr..", ERROR: Can't COPY. Illegal register.")
          if strict then return end
        else
          registers[instr[3]] = stack[stack_ptr]
        end
      elseif instr[2] == 4 then -- push value
        push(instr[3])
      elseif instr[2] == 5 then -- push heap value
        local heap_val = get_heap(stack[stack_ptr])
        if stack_ptr ~= 0 then
          pop()
        else
          print("Line: "..instr_ptr..", ERROR: Can't POP. Empty stack.")
          if strict then return end
        end
        push(heap_val)
      elseif instr[2] == 6 then -- copy value in stack_ptr-1 to heap at addr in stack_ptr
        heap[stack[stack_ptr]] = stack[stack_ptr - 1]
        for i in pairs({0,0}) do
          if stack_ptr ~= 0 then
            pop()
          else
            print("Line: "..instr_ptr..", ERROR: Can't POP. Empty stack.")
            if strict then return end
          end
        end
      end
      
    elseif instr[1] == 2 then -- register math
      if instr[2] == 1 then
        if (instr[3] < 1 or instr[3] > 8) and (instr[4] < 1 or instr[4] > 8) then
          print("Line: "..instr_ptr..", ERROR: Can't ADD. Illegal register.")
          if strict then return end
        else
          push(registers[instr[3]] + registers[instr[4]])
        end 
      elseif instr[2] == 2 then
        if (instr[3] < 1 or instr[3] > 8) and (instr[4] < 1 or instr[4] > 8) then
          print("Line: "..instr_ptr..", ERROR: Can't SUB. Illegal register.")
          if strict then return end
        else
          push(registers[instr[3]] - registers[instr[4]])
        end 
      elseif instr[2] == 3 then
        if (instr[3] < 1 or instr[3] > 8) and (instr[4] < 1 or instr[4] > 8) then
          print("Line: "..instr_ptr..", ERROR: Can't MUL. Illegal register.")
          if strict then return end
        else
          push(registers[instr[3]] * registers[instr[4]])
        end 
      elseif instr[2] == 4 then
        if (instr[3] < 1 or instr[3] > 8) and (instr[4] < 1 or instr[4] > 8) then
          print("Line: "..instr_ptr..", ERROR: Can't DIV. Illegal register.")
          if strict then return end
        else
          push(math.floor((registers[instr[3]] / registers[instr[4]]) + 0.5))
        end 
      end
      
    elseif instr[1] == 3 then -- branch
      if instr[2] == 1 then
        if (instr[3] < 1 or instr[3] > 8) and (instr[4] < 1 or instr[4] > 8) then
          print("Line: "..instr_ptr..", ERROR: Can't BNE. Illegal register.")
          if strict then return end
        else
          if registers[instr[3]] ~= registers[instr[4]] then
            if stack[stack_ptr] ~= nil and instructions[stack[stack_ptr]] ~= nil then
              instr_ptr = stack[stack_ptr]
              pop()
            else
              print("Line: "..instr_ptr..", ERROR: Can't BNE. Location out of bounds.")
              if strict then return end
            end
          end
        end 
      elseif instr[2] == 2 then
        if (instr[3] < 1 or instr[3] > 8) and (instr[4] < 1 or instr[4] > 8) then
          print("Line: "..instr_ptr..", ERROR: Can't BLE. Illegal register.")
          return
        else
          if registers[instr[3]] <= registers[instr[4]] then
            if stack[stack_ptr] ~= nil and instructions[stack[stack_ptr]] ~= nil then
              instr_ptr = stack[stack_ptr]
              pop()
            else
              print("Line: "..instr_ptr..", ERROR: Can't BLE. Location out of bounds.")
              if strict then return end
            end
          end
        end 
      elseif instr[2] == 3 then
        if (instr[3] < 1 or instr[3] > 8) and (instr[4] < 1 or instr[4] > 8) then
          print("Line: "..instr_ptr..", ERROR: Can't BGE. Illegal register.")
          if strict then return end
        else
          if registers[instr[3]] >= registers[instr[4]] then
            if stack[stack_ptr] ~= nil and instructions[stack[stack_ptr]] ~= nil then
              instr_ptr = stack[stack_ptr]
              pop()
            else
              print("Line: "..instr_ptr..", ERROR: Can't BGE. Location out of bounds.")
              if strict then return end
            end
          end
        end 
      end
      
    elseif instr[1] == 4 then -- i/o
      if instr[2] == 1 then
        if stack[stack_ptr] ~= nil then
          io.write(string.format("%c",stack[stack_ptr]))
          pop()
        else
          print("Line: "..instr_ptr..", ERROR: Can't Print_char. Empty stack.")
          if strict then return end
        end
      elseif instr[2] == 2 then
        if stack[stack_ptr] ~= nil then
          io.write(string.format("%d",stack[stack_ptr]))
          pop()
        else
          print("Line: "..instr_ptr..", ERROR: Can't Print_value. Empty stack.")
          if strict then return end
        end
      elseif instr[2] == 3 then
        push(string.byte(io.read(1))) io.read(1) --read the newline
      end
      
    elseif instr[1] == 5 then -- jump
      if stack[stack_ptr] ~= nil  and instructions[stack[stack_ptr]] ~= nil then
        instr_ptr = stack[stack_ptr]
        pop()
      else
        print("Line: "..instr_ptr..", ERROR: Can't JMP. Invalid location.")
        if strict then return end
      end
    end
    
    instr_ptr = instr_ptr + 1
  end
end

--main function
function main()
  --process program into instruction store
  if arg[1] == nil then
    --error out and die the death
    print "ERROR: No program specified"
    return
  else
    --see if we are to use strict execution (default: false)
    if arg[2] and string.match(arg[2], "strict") ~= nil and string.match(arg[2], "false") == nil then
      strict = true
    end
    if arg[3] and string.match(arg[3], "strict") ~= nil and string.match(arg[3], "false") == nil then
      strict = true
    end

    --see if we are to process numbers (default: true)
    if arg[2] and string.match(arg[2], "no_numbers") ~= nil and string.match(arg[2], "false") == nil then
      strict_no_numbers = true
    end
    if arg[3] and string.match(arg[3], "no_numbers") ~= nil and string.match(arg[3], "false") == nil then
      strict_no_numbers = true
    end
  
    process_file(arg[1])
  end

  --interpret program
  interpret()
  io.write("\n") --clean up output
end

--run
main()
